#include <iostream>
#include <sstream>

#include "assert.h"

#include "compile.hh"
#include "symtables.hh"
#include "binding.hh"
#include "basfun.hh"
#include "tristate"

#define CPP_LBOUND_PARAM_NAME(x) (std::string(x)+"_ML")
#define CPP_DIMENSION_PARAM_NAME(x) (std::string(x)+"_MS")
#define CPP_SUBFUNC_NAME(x) (std::string(x)+"_f")
#define CPP_MEMBER_NAME(x) (std::string(x))
#define CPP_UTYPE_NAME(x)  (std::string(x)+"_u")
#define CPP_INT_NAME(x)    (std::string(x)+"_i")
#define CPP_SINGLE_NAME(x) (std::string(x)+"_z")
#define CPP_DOUBLE_NAME(x) (std::string(x)+"_d")
#define CPP_LONG_NAME(x)   (std::string(x)+"_l")
#define CPP_STRING_NAME(x) (std::string(x)+"_s")
#define CPP_ARRAY_NAME(x)  (std::string(x)+"_a")

std::vector<std::string> global_data;

static std::string FunctionReturnVarName;

static const std::pair<std::string, std::string>
    GetTypeBaseNames(const TypeInfo& type, const std::string& name)
{
    switch(type.Type)
    {
        case TypeInfo::Integer: return std::make_pair("BAS_INT", CPP_INT_NAME(name));
        case TypeInfo::Single:  return std::make_pair("BAS_SGN", CPP_SINGLE_NAME(name));
        case TypeInfo::Double:  return std::make_pair("BAS_DBL", CPP_DOUBLE_NAME(name));
        case TypeInfo::Long:    return std::make_pair("BAS_LNG", CPP_LONG_NAME(name));
        case TypeInfo::String:
            return type.fixlength
                ? /* FIXME */
                  std::make_pair("BAS_STR", CPP_STRING_NAME(name))
                : std::make_pair("BAS_STR", CPP_STRING_NAME(name));
        case TypeInfo::UserType:
            return std::make_pair(type.utypename, CPP_UTYPE_NAME(name));
    }
    /* not reached, but keeps compiler happy */
    assert(!"not_reached");
    return std::make_pair(type.utypename, CPP_UTYPE_NAME(name));
}

enum StrCompileDefs
{
    StrCompileMultipurpose,
    StrCompilePrinting,
    StrCompileConstOnly
};
static const std::string CompileStringConst
   (const std::string&str,
    StrCompileDefs defs = StrCompileMultipurpose)
{
    char esc = '"';
    if(defs == StrCompilePrinting && str.size() == 1)
    {
        esc = '\'';
    }
    std::string result;
    result += esc;
    bool contains_nul_char = false;
    for(unsigned a=0; a<str.size(); ++a)
    {
        unsigned char c = str[a];
        if(c=='\n') { result += "\\n"; continue; }
        if(c=='\r') { result += "\\r"; continue; }
        if(c=='\t') { result += "\\t"; continue; }
        if(c=='\b') { result += "\\b"; continue; }
        if(c=='\a') { result += "\\a"; continue; }
        if(c==esc || c=='\\') { result += '\\'; result += str[a]; continue; }
        if(c < 32 || c >= 127)
        {
            if(c == 0) contains_nul_char = true;
            char Buf[64];
            int width = 1;
            if(a+1 < str.size() && str[a+1]>='0' && str[a+1]<='7') width=3;
            sprintf(Buf, "\\%0*o", width, c);
            result += Buf;
            continue;
        }
        result += str[a];
    }
    result += esc;
    if(esc == '"' && defs != StrCompileConstOnly) // && contains_nul_char)
    {
        /* Always create an explicit string, because otherwise
         * we will fail when stringconsts are handled together
         * in a comparison for example
         */
        std::stringstream res2;
        res2 << "std::string(" << result << ", " << str.size() << ")";
        result = res2.str();
    }
    return result;
}

enum VarDeclType { Init_Var, Param_Ref, Param_Copy, Param_Const };

static struct
{
    std::string resulting_name;
    
    void operator() (std::ostream& out, const std::string& name,
                     TypeP typep, VarDeclType decl_type = Init_Var
                    )
    {
        assert(typep);
        const TypeInfo& type = *typep;

        const std::pair<std::string, std::string> base = GetTypeBaseNames(type, name);

        switch(type.arraytype)
        {
            case TypeInfo::NotArray:
            {
                bool const_ref = false;
                if(decl_type == Param_Const)
                    const_ref = !type.is_num();

                if(const_ref) out << "const ";
                out << base.first << ' ';
                if(decl_type == Param_Ref || const_ref) out << "&";
                out << (resulting_name=base.second);
                break;
            }
            case TypeInfo::StaticArray:
            {
                out << base.first << ' ';
                if(decl_type == Param_Copy)
                {
                    fprintf(stderr, "Warning: BYVAL Array parameters not supported\n");
                }
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                for(unsigned a=0; a<type.dimensions.size(); ++a)
                    out << (!a ? "[" : "][") << type.dimensions[a].second;
                out << "]";
                break;
            }
            case TypeInfo::DynamicArray:
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_DYNARRAY<" << base.first << ", " << type.dimensions.size() << "> ";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << "&";
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                
                if(decl_type == Init_Var)
                {
                    for(unsigned a=0; a<type.dimensions.size(); ++a)
                    {
                        out << (!a ? "(" : ", ")
                            << type.dimensions[a].first
                            << ","
                            << type.dimensions[a].second;
                    }
                    out << ')';
                }
                break;
            }
            case TypeInfo::ParamArray:
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_MAGICARRAY<" << base.first << ">";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << "&";
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                break;
            }
        }
    }
private:
} CompileVarDeclaration;

static void CompileTypeDefs(const Scope& scope)
{
    for(hash_map<std::string, UtypeP>::const_iterator
        i = scope.types.begin();
        i != scope.types.end();
        ++i)
    {
        UtypeP p = i->second;
        assert(p);
        
        std::cout << "struct " << i->first << " {\n";
        
        for(unsigned a=0; a<p->size(); ++a)
        {
            std::cout << "  ";
            CompileVarDeclaration(std::cout, CPP_MEMBER_NAME((*p)[a].first), (*p)[a].second );
            std::cout << ";\n";
        }
        
        std::cout << "};\n";
    }
}

static void CompileVars(const ScopeVarListType& vars, tristate shared_only, bool implicit_static=false)
{
    for(ScopeVarListType::const_iterator i = vars.begin(); i != vars.end(); ++i)
    {
        ctxp ctx = i->second;
        assert(ctx);
        if(!shared_only.is_maybe() && ctx->is_shared != shared_only) continue;
        if(ctx->is_param) continue;

        if(ctx->is_static || implicit_static) std::cout << "static ";
        
        CompileVarDeclaration(std::cout, i->first, ctx->type);
        std::cout << ";\n";
    }
}

static void CompileVars(const Scope& scope, tristate shared_only, bool implicit_static=false)
{
    CompileVars(scope.Utype  .scalars, shared_only, implicit_static);
    CompileVars(scope.Utype  .arrays, shared_only, implicit_static);
    CompileVars(scope.Int    .scalars, shared_only, implicit_static);
    CompileVars(scope.Int    .arrays, shared_only, implicit_static);
    CompileVars(scope.Single .scalars, shared_only, implicit_static);
    CompileVars(scope.Single .arrays, shared_only, implicit_static);
    CompileVars(scope.Double .scalars, shared_only, implicit_static);
    CompileVars(scope.Double .arrays, shared_only, implicit_static);
    CompileVars(scope.Long   .scalars, shared_only, implicit_static);
    CompileVars(scope.Long   .arrays, shared_only, implicit_static);
    CompileVars(scope.String .scalars, shared_only, implicit_static);
    CompileVars(scope.String .arrays, shared_only, implicit_static);
}

static void CompileSubFunc(const std::string& name, const SubRoutine& sub, bool decl_only)
{
    std::cout << "static ";
    if(sub.returntype)
    {
        const std::pair<std::string, std::string> base
            = GetTypeBaseNames(*sub.returntype, name);
        /* no "const" here, to not hinder passing the result as reference */
        std::cout << base.first;
    }
    else
        std::cout << "void";

    std::cout << " " << CPP_SUBFUNC_NAME(name) << "(";
    
    for(unsigned a=0; a<sub.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& param = sub.params[a];
        if(a) std::cout << ", ";
        CompileVarDeclaration(std::cout,
                              param.name,
                              param.type,
                              param.was_modified
                               ? param.passconvention ==
                                   SubRoutine::ParamSpec::ByVal
                                       ? Param_Copy
                                       : Param_Ref
                               : Param_Const);
        if(param.is_array)
        {
            std::cout << ", BAS_INT " << CPP_LBOUND_PARAM_NAME(param.name);
            std::cout << ", BAS_INT " << CPP_DIMENSION_PARAM_NAME(param.name);
        }
    }
    
    std::cout << ')';
    
    if(decl_only)
        std::cout << ";\n";
    else
    {
        std::cout << "\n";
    }
}

static void CompileSubDefs(const Scope& scope)
{
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope.subroutines.begin(); i != scope.subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        CompileSubFunc(i->first, *sub, true);
    }
}

static void CompileCode(stmtp codep, const Scope& scope)
{
    if(!codep) return;
    CompileCode(codep->get_prev(), scope);
    
    if(codep->MayErr()) std::cout << "B_ERROR_BEGIN() ";
    codep->Compile(scope);
    if(codep->MayErr()) std::cout << " B_ERROR_OVER() ";
}

static const std::string GenerateLabelName(long double lab, const Scope& scope)
{
    if(scope.labels.find(lab) == scope.labels.end())
    {
        fprintf(stderr, "Error: Label not defined in current scope\n");
        return "L??";
    }
    unsigned lnum = 0;
    static std::map<std::pair<const Scope*, long double>, unsigned> ldata;
    std::pair<const Scope*, long double> seed(&scope, lab);
    std::map<std::pair<const Scope*, long double>, unsigned>::const_iterator
        l = ldata.find(seed);
    if(l != ldata.end()) lnum = l->second;
    else lnum = ldata[seed] = ldata.size() + 1;
    std::stringstream tmp; tmp << 'L' << lnum;
    return tmp.str();
}

static void CompileExpr(exprp e, const Scope& scope,
                        ExprCompileDefs defs=exprdef_nothingparticular)
{
    if(!e) return;
    
    if(defs == exprdef_cast_to_int)
    {
        TypeP type = e->get_type();
        if(type && type->is_num())
        {
            if(type->Type != TypeInfo::Integer
            && type->Type != TypeInfo::Long)
            {
                std::cout << "BasCLNG(";
                defs = exprdef_nothingparticular;
                e->Compile(scope, defs);
                std::cout << ")";
                return;
            }
        }
    }
    e->Compile(scope, defs);
}

unsigned TranslateLabelToDataPtr(const long double l)
{
    // TODO: Use current scope and parent scope
    return 0;
}

static void VerifyLabels(const Scope& scope)
{
    for(std::map<long double, LabelDef>::const_iterator
        i = scope.labels.begin(); i != scope.labels.end(); ++i)
    {
        if(!i->second.defined)
            fprintf(stderr, "Error: Label %Lg not defined\n", i->first);
    }
}

/*
PLAN:
    Define typedefs. - DONE
    Define global variables. -SHARED ONLY DONE
    Declare subroutines. -DONE
    
    Compile code.
    
    Compile each function.

NOTES:
    Constants are not defined. They are compiled!

TODO:
    SHARED imports?
    
*/
void Compile(stmtp main_codep,
             const std::vector<std::string>& global_data)
{
    ScopeP scope = GetRootScope();
    VerifyLabels(*scope);
    
    std::cout << "static const unsigned DATA_COUNT = " << global_data.size() << ";\n";
    if(!global_data.empty())
    {
        std::cout << "static unsigned DATA_PTR = 0;\n";
        std::cout << "static const char* const DATA_TABLE[DATA_COUNT] = {\n";
        for(unsigned a=0; a<global_data.size(); ++a)
        {
            //std::cout << "// " << a << "\n";
            std::cout << CompileStringConst(global_data[a], StrCompileConstOnly);
            if(a+1 < global_data.size()) std::cout << ",";
            std::cout << "\n";
        }
        std::cout << "}\n";
    }
    
    CompileTypeDefs(*scope);
    CompileVars(*scope, true, true); // shared only
    CompileSubDefs(*scope);

    std::cout << "int main(int argc, char**argv)\n";
    std::cout << "{\n";
    CompileVars(*scope, false, false); // local only
    std::cout << "B_ERR_HEADER();\n";
    CompileCode(main_codep, *scope);
    std::cout << "}\n";

    /* Compile subs */
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope->subroutines.begin(); i != scope->subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        ScopeP subscope = sub->sub_scope;
        VerifyLabels(*subscope);
        
        CompileSubFunc(i->first, *sub, false);
        std::cout << "{\n";
        
        if(sub->returntype)
        {
            CompileVarDeclaration(std::cout, "FunctionReturn", sub->returntype);
            FunctionReturnVarName = CompileVarDeclaration.resulting_name;
            std::cout << ";\n";
        }
        
        CompileVars(*subscope, false, sub->is_static); // local only
        CompileCode(sub->code, *subscope);
        
        std::cout << "subfunc_exit:; \n";

        if(sub->returntype)
        {
            std::cout << "    return " << FunctionReturnVarName << ";\n";
        }
        std::cout << "}\n";
    }
}

void stmt_if::Compile(const Scope& scope)
{
    if(!thenp && elsep && elseifp.empty())
    {
        std::cout << "if(!("; CompileExpr(condition, scope, exprdef_pure_condition);
        std::cout << ")) {\n"; CompileCode(elsep, scope);
        std::cout << "}\n";
    }
    else
    {
        std::cout << "if("; CompileExpr(condition, scope, exprdef_pure_condition);
        std::cout << ") {\n";
            if(thenp) CompileCode(thenp, scope);
        std::cout << "}";
        for(std::list<std::pair<exprp, stmtp> >::const_iterator
            i = elseifp.begin(); i != elseifp.end(); ++i)
        {
            std::cout << " else if("; CompileExpr(i->first, scope, exprdef_pure_condition);
            std::cout << ") {\n";
            if(i->second) CompileCode(i->second, scope);
            std::cout << "}";
        }
        if(elsep)
        {
            std::cout << " else {\n";
            CompileCode(elsep, scope);
            std::cout << "}";
        }
        std::cout << "\n";
    }
}
void stmt_doloop::Compile(const Scope& scope)
{
    std::cout << "{ __label__ do_exit;\n";
    if(precondition)
    {
        std::cout << "while("; CompileExpr(precondition, scope, exprdef_pure_condition); std::cout << ")\n{\n";
    }
    else if(!postcondition)
        std::cout << "for(;;)\n{\n";
    else
        std::cout << "do {\n";

    CompileCode(bodyp, scope);
    
    if(postcondition)
    {
        if(precondition)
        {
            std::cout << "if("; CompileExpr(postcondition, scope, exprdef_pure_condition);
            std::cout << ") break;";
        }
        else
        {
            std::cout << "} while("; CompileExpr(postcondition, scope, exprdef_pure_condition);
            std::cout << ");";
        }
    }
    else
        std::cout << "}";
    std::cout << " do_exit:; }\n";
}
void stmt_for::Compile(const Scope& scope)
{
    static unsigned ForCounter=0;
    char ForTempName[64]; sprintf(ForTempName, "FORTEMP%u", ForCounter++);
    
    bool nonconst_step = true;
    QBtoken step_operator = (QBtoken)'<';
    
    if(!step)
        nonconst_step = false;
    else
    {
        if(expr_number* num = dynamic_cast<expr_number*> (&*step))
        {
            nonconst_step = false;
            if(num < 0) step_operator = (QBtoken)'>';
        }
    }
    
    std::pair<std::string, std::string> ForTemp;
    if(nonconst_step)
    {
        ForTemp = GetTypeBaseNames(*step->get_type(), ForTempName);
    }
    
    std::cout << "{ __label__ for_exit;\n";
    if(nonconst_step)
    {
        std::cout << "const " << ForTemp.first << " " << ForTempName << " = ";
        CompileExpr(step, scope);
        std::cout << ";\n";
    }
    std::cout << "for(";
        stmt_let(var, begin).Compile(scope); // it appends a semicolon, too.
        if(nonconst_step)
        {
            std::cout << "(" << ForTempName << " < 0 ? (";
            //negative step?
                CompileExpr(var, scope); std::cout << " > "; CompileExpr(end, scope);
            std::cout << ") : (";
            //positive step?
                CompileExpr(var, scope); std::cout << " < "; CompileExpr(end, scope);
            std::cout << ')';
        }
        else
        {
            expr_binary(step_operator, var, end).Compile(scope, exprdef_pure_condition);
        }
        
        std::cout << "; ";
        
        CompileExpr(var, scope);
        if(!step)
            { std::cout << " += 1 "; }
        else if(nonconst_step)
            { std::cout << " += " << ForTempName; }
        else
            { std::cout << " += "; CompileExpr(step, scope); }
    std::cout << ")\n{\n";
        CompileCode(bodyp, scope);
    std::cout << "} for_exit:; }\n";
}
void stmt_let::Compile(const Scope& scope)
{
    CompileExpr(target, scope);
    std::cout << " = ";
    CompileExpr(value, scope);
    std::cout << ";\n";
}

static void CompileCaseTest(exprp expr, const std::string& varname,
                            const Scope& scope,
                            bool IsSwitchCase)
{
    ExprCompileDefs defs = IsSwitchCase ? exprdef_cast_to_int : exprdef_nothingparticular;
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(&*expr))
    {
        CompileCaseTest(e->a, varname, scope, defs);
        if(!IsSwitchCase) std::cout << " || ";
        CompileCaseTest(e->b, varname, scope, defs);
    }
    else if(expr_caseis* e = dynamic_cast<expr_caseis*>(&*expr))
    {
        std::string op = "==";
        switch(e->token)
        {
            case (QBtoken)'>':    op = ">"; break;
            case (QBtoken)'<':    op = "<"; break;
            case (QBtoken)'=':    op = "=="; break;
            case NOTEQUAL:        op = "!="; break;
            case LESSOREQUAL:     op = "<="; break;
            case GREATEROREQUAL:  op = ">="; break;
            default: assert(!"unexpected caseis token");
        }
        if(IsSwitchCase && op == "==") op="";
        if(IsSwitchCase) std::cout << "case " << op;
        else std::cout << varname << ' ' << op << ' ';
        CompileExpr(e->ref, scope, defs);
        if(IsSwitchCase) std::cout << ":\n";
    }
    else if(expr_caserange* e = dynamic_cast<expr_caserange*>(&*expr))
    {
        if(IsSwitchCase)
        {
            std::cout << "case ";
            CompileExpr(e->min, scope, defs);
            std::cout << " ... ";
            CompileExpr(e->max, scope, defs);
            std::cout << ":\n";
        }
        else
        {
            std::cout << "(" << varname << " >= ";
            CompileExpr(e->min, scope, defs);
            std::cout << " && " << varname << " <= ";
            CompileExpr(e->max, scope, defs);
            std::cout << ')';
        }
    }
    else assert(!"unexpected case type");
}

void stmt_select::Compile(const Scope& scope)
{
    if(IsEntirelyIntegerCase())
    {
        std::cout << "switch(";
        CompileExpr(ref, scope, exprdef_cast_to_int);
        std::cout << ")\n{\n";
        for(std::list<
            std::pair<exprp, stmtp>
                 >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        {
            if(expr_caseelse* e = dynamic_cast<expr_caseelse*>(&*i->first))
            {
                std::cout << "default:\n";
                CompileCode(i->second, scope);
                break;
            }
            CompileCaseTest(i->first, "?", scope, true);
            CompileCode(i->second, scope);
            std::cout << "break;\n";
        }
        std::cout << "}\n";
    }
    else
    {
        static unsigned SwitchCounter=0;
        char SwitchTempName[64]; sprintf(SwitchTempName, "SWITCHTEMP%u", SwitchCounter++);
        std::pair<std::string, std::string> SwitchTemp;
        SwitchTemp = GetTypeBaseNames(*ref->get_type(), SwitchTempName);
        
        bool first = true;
        std::cout << "{ const " << SwitchTemp.first << " " << SwitchTemp.second
                  << " = "; CompileExpr(ref, scope);
        std::cout << ";\n";
        
        for(std::list<
            std::pair<exprp, stmtp>
                 >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        {
            if(!first) std::cout << "else ";
            if(expr_caseelse* e = dynamic_cast<expr_caseelse*>(&*i->first))
            {
                std::cout << "{\n";
                CompileCode(i->second, scope);
                std::cout << "} ";
                break;
            }
            std::cout << "if(";
            CompileCaseTest(i->first, SwitchTemp.second, scope, false);
            std::cout << ")\n{\n";
            CompileCode(i->second, scope);
            std::cout << "} ";
            first=false;
        }
    
        std::cout << "}\n";
    }
}

void stmt_print::Compile(const Scope& scope)
{
    static unsigned UsingCounter=0;
    char TempName[64]; sprintf(TempName, "PTEMP%u", UsingCounter++);
    std::string OutTempName="OUT"+std::string(TempName),
              UsingTempName="USING"+std::string(TempName);

    std::cout << "{ PrintStreamBase& " << OutTempName << " = ";
    if(target)
    {
        std::cout << "GET_STREAM_OUTPUT("; CompileExpr(target, scope);
        std::cout << ')';
    }
    else if(is_lprint)
    {
        std::cout << "GET_PRINTER()";
    }
    else
    {
        std::cout << "GET_STDOUT()";
    }
    std::cout << ";\n";
    
    if(using_string)
    {
        std::cout << "UsingFormatter " << UsingTempName << "(";
        CompileExpr(using_string, scope);
        std::cout << ");\n";
    }
    
    bool outputting = false;
    
    for(std::list<std::pair<QBtoken, exprp> >::const_iterator
        i = params.begin(); i != params.end(); ++i)
    {
        switch(i->first)
        {
            case (QBtoken)0:
                if(using_string)
                {
                    if(outputting) { std::cout << ";\n"; outputting=false; }
                    std::cout << UsingTempName << ".DoParam(" << OutTempName
                              << ", ";
                    CompileExpr(i->second, scope);
                    std::cout << ");\n";
                }
                else
                {
                    if(!outputting) { outputting=true; std::cout << OutTempName; }
                    std::cout << " << ";
                    CompileExpr(i->second, scope, exprdef_may_produce_char_from_string);
                }
                break;
            case SPC:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                std::cout << OutTempName << ".DoSPC("; CompileExpr(i->second, scope);
                std::cout << ");\n";
                break;
            case TAB:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                if(i->second)
                {
                    std::cout << OutTempName << ".DoTAB("; CompileExpr(i->second, scope);
                    std::cout << ");\n";
                }
                else
                {
                    std::cout << OutTempName << ".DoNextTAB();\n";
                }
                break;
            case CRLF:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                std::cout << OutTempName << ".DoCRLF();\n";
                break;
            default: assert(!"unexpected print operator");
        }
    }
    
    if(outputting) { std::cout << "; "; outputting=false; }
    std::cout << "}\n";
}

void stmt_input::Compile(const Scope& scope)
{
    static unsigned UsingCounter=0;
    char TempName[64]; sprintf(TempName, "PTEMP%u", UsingCounter++);
    std::string InTempName="IN"+std::string(TempName);

    std::cout << "{ InputStreamBase& " << InTempName << " = ";
    if(source)
    {
        std::cout << "GET_STREAM_INPUT("; CompileExpr(source, scope);
        std::cout << ')';
    }
    else
    {
        std::cout << "GET_STDIN()";
    }
    std::cout << ";\n";

    if(!is_linput && !source && !prompt) prompt = new expr_string("? ");
    
    if(!source && !is_linput)  std::cout << "for(;;) {";
    
    if(!source && prompt)
        { std::cout << "GET_STDOUT() << ";
          CompileExpr(prompt, scope, exprdef_may_produce_char_from_string);
          std::cout << ";\n"; }
    
    for(std::list<exprp>::const_iterator
        i = params.begin(); i != params.end(); ++i)
    {
        std::string method = is_linput ? "ReadLine" : "InputTo";
        std::cout << InTempName << '.' << method
                  << '('; CompileExpr(*i,scope); std::cout << ");\n";
    }
    
    /* Input error from console results in "?Redo from start"
     * Input error from file results in error 62 */
    
    if(!source && !is_linput)
        std::cout << "if(" << InTempName << ".Satisfied()) break; \n"
                  << "GET_STDOUT() << " << InTempName << ".GetRedoMsg(); }\n";
    else
        std::cout << "if(!" << InTempName << ".Satisfied()) B_ERROR(62);\n";
    
    if(!source && !keep_cursor)
        std::cout << "GET_STDOUT().DoCRLF();\n";
        
    std::cout << "}\n";
}



void stmt_return::Compile(const Scope& scope)
{
    switch(token)
    {
        case EXITDO: std::cout << "goto do_exit;\n"; break;
        case EXITFOR: std::cout << "goto for_exit;\n"; break;
        case END: std::cout << "B_END();\n"; break;
        case STOP: std::cout << "B_STOP();\n"; break;
        case EXITSUB:
        case EXITFUNCTION:
        case EXITDEF: std::cout << "goto subfunc_exit;\n"; break;
        case RETURN:
        {
            if(label)
            {
                expr_label* e;
                assert((e = dynamic_cast<expr_label*> (&*label)) && e);
                long double lab = e->get_label();
                std::string lname = GenerateLabelName(lab, scope);
                std::cout << "B_RETURN_TO(" << lname << ");\n";
            }
            else
            {
                std::cout << "B_RETURN();\n";
            }
            break;
        }
        case RESUME:
            // Note: May contain an explicit label
            if(label)
            {
                expr_label* e;
                assert((e = dynamic_cast<expr_label*> (&*label)) && e);
                long double lab = e->get_label();
                std::string lname = GenerateLabelName(lab, scope);
                std::cout << "B_RESUME_TO(" << lname << ");\n";
            }
            else
            {
                std::cout << "B_RESUME();\n";
            }
            break;
        case RESUMENEXT:
            std::cout << "B_RESUME_NEXT();\n";
            break;
        default: assert(!"unexpected stmt_return token");
    }
}
void stmt_goto::Compile(const Scope& scope)
{
    expr_label* e = 0;
    long double lab=0;
    std::string lname;
    if(label)
    {
        e = dynamic_cast<expr_label*> (&*label);
        assert(e);
        lab = e->get_label();
        lname = GenerateLabelName(lab, scope);
    }
    switch(token)
    {
        case GOTO:
            std::cout << "goto " << lname << ";\n";
            break;
        case GOSUB:
            std::cout << "B_GOSUB(" << lname << ");\n";
            break;
        case ON_ERROR:
            std::cout << "ON_ERROR_GOTO(" << lname << ");\n";
            break;
        case RESTORE:
            std::cout << "B_RESTORE(";
            if(!label) std::cout << '0';
            else
            {
                std::map<long double, LabelDef>::const_iterator
                    i = scope.labels.find(lab);
                if(i != scope.labels.end())
                    std::cout << i->second.dataptr;
                
                std::cout << " /* " << lname << " */";
            }
            std::cout << ");\n";
            break;
        default: assert(!"unexpected stmt_goto token");
    }
}
void stmt_label::Compile(const Scope& scope)
{
    long double lab = label;
    std::string lname = GenerateLabelName(lab, scope);
    std::cout << lname << ": ;\n";
}


void stmt_subcall::Compile(const Scope& scope)
{
    exprp e = new expr_userfun(name, 0, params);
    CompileExpr(e, scope);
    std::cout << ";\n";
}
void stmt_read::Compile(const Scope& scope)
{
    std::cout << "B_READ("; CompileExpr(target, scope);
    std::cout << ");\n";
}

void stmt_fileopen::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_redim::Compile(const Scope& scope) { stmt::Compile(scope); }



void expr_string::Compile(const Scope& scope, ExprCompileDefs defs)
{
    std::cout << CompileStringConst
        (this->value,
           defs==exprdef_may_produce_char_from_string
           ? StrCompilePrinting
           : StrCompileMultipurpose);
}
void expr_number::Compile(const Scope& scope, ExprCompileDefs defs)
{
    if(is_int())
        std::cout << get_i();
    else
        std::cout << get_f();
}
void expr_unary::Compile(const Scope& scope, ExprCompileDefs defs)
{
    std::cout << "(";
    ExprCompileDefs subdefs;
    switch(token)
    {
        case NOT:
            if(defs == exprdef_pure_condition) { std::cout << "!"; subdefs=defs; break; }
            std::cout << "~";
            break;
        case (QBtoken)'-':
            std::cout << "-";
            break;
        default: assert(!"unexpected expr_unary token");
    }
    CompileExpr(p, scope, subdefs);
    std::cout << ')';
}
void expr_binary::Compile(const Scope& scope, ExprCompileDefs defs)
{
    std::string fucast, funame, cast0, op, cast1, finish;
    ExprCompileDefs subdefs = exprdef_nothingparticular;

    switch(token)
    {
        case XOR:
            op = "^";
        FinishInteger:
            subdefs = exprdef_cast_to_int;
            break;
        case OR:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "||"; subdefs=defs; break; }
            op = "|"; goto FinishInteger;
        case AND:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "&&"; subdefs=defs; break; }
            op = "&"; goto FinishInteger;
        case (QBtoken)'>':
            op = ">";
        FinishCompare:
            if(defs == exprdef_pure_condition)
                subdefs = defs;
            else
            {
                fucast = "(";
                finish = ") ? -1 : 0";
            }
            break;
        case (QBtoken)'<':    op = "<"; goto FinishCompare;
        case (QBtoken)'=':    op = "=="; goto FinishCompare;
        case NOTEQUAL:        op = "!="; goto FinishCompare;
        case LESSOREQUAL:     op = "<="; goto FinishCompare;
        case GREATEROREQUAL:  op = ">="; goto FinishCompare;
        case (QBtoken)'+':    op = "+"; break;
        case (QBtoken)'-':    op = "-"; break;
        case (QBtoken)'\\':   op = "/"; subdefs = exprdef_cast_to_int; break;
        case (QBtoken)'*':    op = "*"; break;
        case (QBtoken)'/':    op = "/"; cast1 = "(BAS_DBL)"; break;
        case (QBtoken)'^':    funame = "std::pow("; op=","; finish=")"; break;
        case MOD:             funame = "BasMODop("; op=","; finish=")"; break;
        default: assert(!"unexpected expr_binary token");
    }
    std::cout << fucast << funame << "("
              << cast0 << "("; CompileExpr(p0, scope, subdefs); std::cout << ')'
              << ' ' << op << ' '
              << cast1 << "("; CompileExpr(p1, scope, subdefs); std::cout << ')'
              << finish
              << ')';
}

void expr_userfun::Compile(const Scope& scope, ExprCompileDefs defs)
{
    std::string basename = BasUCASE(StripTypeSuffix(name));
    std::cout << CPP_SUBFUNC_NAME(basename) << '(';
    for(unsigned a=0; a<params.size(); ++a)
    {
        if(a) std::cout << ", ";
        CompileExpr(params[a], scope, exprdef_nothingparticular);
    }
    std::cout << ')';
}

void expr_builtin_numeric::Compile(const Scope& scope, ExprCompileDefs defs)
{ expr::Compile(scope, exprdef_nothingparticular); }
void expr_builtin_string::Compile(const Scope& scope, ExprCompileDefs defs)
{ expr::Compile(scope, exprdef_nothingparticular); }
void expr_lbound::Compile(const Scope& scope, ExprCompileDefs defs)
{ expr::Compile(scope, exprdef_nothingparticular); }
void expr_ubound::Compile(const Scope& scope, ExprCompileDefs defs)
{ expr::Compile(scope, exprdef_nothingparticular); }

// SPECIALS
void expr_label::Compile(const Scope& scope, ExprCompileDefs defs)    { assert(!"not reached"); }
void expr_caseelse::Compile(const Scope& scope, ExprCompileDefs defs) { assert(!"not reached"); }
void expr_caseis::Compile(const Scope& scope, ExprCompileDefs defs)   { assert(!"not reached"); }
void expr_caserange::Compile(const Scope& scope, ExprCompileDefs defs){ assert(!"not reached"); }
void expr_caseor::Compile(const Scope& scope, ExprCompileDefs defs)   { assert(!"not reached"); }

// LVALUES
void expr_var::Compile(const Scope& scope, ExprCompileDefs defs)
{
    assert(type);
    
    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*type, BasUCASE(StripTypeSuffix(name)));

    if(defs==exprdef_membername)
        base.second = CPP_MEMBER_NAME(base.second);
    
    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
            std::cout << base.second;
            break;
        case TypeInfo::StaticArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            for(unsigned a=0; a<indices.size(); ++a)
            {
                std::cout << '[';
                CompileExpr(indices[a], scope, exprdef_cast_to_int);
                std::cout << ']';
            }
            break;
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            std::cout << ".Index(";
            for(unsigned a=0; a<indices.size(); ++a)
            {
                if(a > 0) std::cout << ", ";
                CompileExpr(indices[a], scope, exprdef_cast_to_int);
            }
            std::cout << ')';
            break;
    }
    if(hasmember())
    {
        std::cout << '.';
        member->Compile(scope, exprdef_membername); // not CompileExpr here
    }
}
void expr_returnvalueholder::Compile(const Scope& scope, ExprCompileDefs defs)
{
    std::cout << FunctionReturnVarName;
}
